@@ -1,68 +1,54 @@
-/*
-Copyright (c) 2009-2014 Sony Pictures Imageworks Inc., et al.
-All Rights Reserved.
+// Copyright Contributors to the Open Shading Language project.
+// SPDX-License-Identifier: BSD-3-Clause
+// https://github.com/AcademySoftwareFoundation/OpenShadingLanguage
 
-Redistribution and use in source and binary forms, with or without
-modification, are permitted provided that the following conditions are
-met:
-* Redistributions of source code must retain the above copyright
-  notice, this list of conditions and the following disclaimer.
-* Redistributions in binary form must reproduce the above copyright
-  notice, this list of conditions and the following disclaimer in the
-  documentation and/or other materials provided with the distribution.
-* Neither the name of Sony Pictures Imageworks nor the names of its
-  contributors may be used to endorse or promote products derived from
-  this software without specific prior written permission.
-THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
-A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-*/
+// clang-format off
 
 #pragma once
 
 
-#include "oslconfig.h"
-
-#include <OpenImageIO/ustring.h>
+#include <OSL/oslconfig.h>
 
 
 OSL_NAMESPACE_ENTER
 
 class RendererServices;
+template<int WidthT>
+class BatchedRendererServices;
 class ShadingContext;
 struct ShaderGlobals;
 
+// Tags for polymorphic dispatch
+template <int SimdWidthT>
+class WidthOf {};
 
 
 /// Opaque pointer to whatever the renderer uses to represent a
 /// (potentially motion-blurred) coordinate transformation.
-typedef const void * TransformationPtr;
+typedef const void* TransformationPtr;
 
 
 // Callbacks for closure creation
-typedef void (*PrepareClosureFunc)(RendererServices *, int id, void *data);
-typedef void (*SetupClosureFunc)(RendererServices *, int id, void *data);
+typedef void (*PrepareClosureFunc)(RendererServices*, int id, void* data);
+typedef void (*SetupClosureFunc)(RendererServices*, int id, void* data);
 
+// Turn off warnings about unused params for this file, since we have lots
+// of declarations with stub function bodies.
+OSL_PRAGMA_WARNING_PUSH
+OSL_GCC_PRAGMA(GCC diagnostic ignored "-Wunused-parameter")
 
 
 
 /// RendererServices defines an abstract interface through which a
 /// renderer may provide callback to the ShadingSystem.
 class OSLEXECPUBLIC RendererServices {
+    // Keep interface in sync with rs_free_function.h
 public:
     typedef TextureSystem::TextureHandle TextureHandle;
     typedef TextureSystem::Perthread TexturePerthread;
 
 
-    RendererServices (TextureSystem *texsys=NULL);
+    RendererServices(TextureSystem* texsys = NULL);
     virtual ~RendererServices () { }
 
     /// Given the name of a 'feature', return whether this RendererServices
@@ -82,7 +68,7 @@ public:
     /// transformation at the given time.  Return true if ok, false
     /// on error.
     virtual bool get_matrix (ShaderGlobals *sg, Matrix44 &result,
-                             TransformationPtr xform, float time) = 0;
+                             TransformationPtr xform, float time) { return false; }
 
     /// Get the 4x4 matrix that transforms by the specified
     /// transformation at the given time.  Return true if ok, false on
@@ -97,7 +83,7 @@ public:
     /// time value is given, also return false if the transformation may
     /// be time-varying.
     virtual bool get_matrix (ShaderGlobals *sg, Matrix44 &result,
-                             TransformationPtr xform) = 0;
+                             TransformationPtr xform) { return false; }
 
     /// Get the 4x4 matrix that transforms by the specified
     /// transformation.  Return true if ok, false on error.  Since no
@@ -112,7 +98,7 @@ public:
     /// 'from' coordinate system to "common" space at the given time.
     /// Returns true if ok, false if the named matrix is not known.
     virtual bool get_matrix (ShaderGlobals *sg, Matrix44 &result,
-                             ustring from, float time) = 0;
+                             ustring from, float time) { return false; }
 
     /// Get the 4x4 matrix that transforms points from "common" space to
     /// the named 'to' coordinate system to at the given time.  The
@@ -127,7 +113,7 @@ public:
     /// transformation may be time-varying (as well as if it's not found
     /// at all).
     virtual bool get_matrix (ShaderGlobals *sg, Matrix44 &result,
-                             ustring from) = 0;
+                             ustring from) { return false; }
 
     /// Get the 4x4 matrix that transforms points from "common" space to
     /// the named 'to' coordinate system.  Since there is no time value
@@ -147,7 +133,7 @@ public:
     /// the name of one of the coordinate systems).  A renderer is free
     /// to not implement this, in which case the default implementation
     /// is simply to make appropriate calls to get_matrix and
-    /// get_inverse_matrix.  The existance of this method is to allow
+    /// get_inverse_matrix.  The existence of this method is to allow
     /// some renderers to provide transformations that cannot be
     /// expressed by a 4x4 matrix.
     ///
@@ -183,36 +169,34 @@ public:
     /// otherwise just fail by returning 'false'.
     virtual bool get_attribute (ShaderGlobals *sg, bool derivatives,
                                 ustring object, TypeDesc type, ustring name,
-                                void *val ) = 0;
+                                void *val) { return false; }
 
     /// Similar to get_attribute();  this method will return the 'index'
     /// element of an attribute array.
     virtual bool get_array_attribute (ShaderGlobals *sg, bool derivatives,
                                       ustring object, TypeDesc type,
-                                      ustring name, int index, void *val ) = 0;
+                                      ustring name, int index, void *val) { return false; }
 
     /// Get the named user-data from the current object and write it into
     /// 'val'. If derivatives is true, the derivatives should be written into val
     /// as well. Return false if no user-data with the given name and type was
     /// found.
     virtual bool get_userdata (bool derivatives, ustring name, TypeDesc type,
-                               ShaderGlobals *sg, void *val) = 0;
-
-    /// Does the current object have the named user-data associated with it?
-    virtual bool has_userdata (ustring name, TypeDesc type, ShaderGlobals *sg) = 0;
+                               ShaderGlobals *sg, void *val) { return false; }
 
     /// Given the name of a texture, return an opaque handle that can be
     /// used with texture calls to avoid the name lookups.
-    virtual TextureHandle * get_texture_handle (ustring filename);
+    virtual TextureHandle * get_texture_handle (ustring filename,
+                                                ShadingContext *context);
 
     /// Return true if the texture handle (previously returned by
     /// get_texture_handle()) is a valid texture that can be subsequently
     /// read or sampled.
     virtual bool good (TextureHandle *texture_handle);
 
-    /// Return a thread-specific opaque pointer to the texture system.
-    /// Knowing the context may help this go faster.
-    virtual TexturePerthread * get_texture_perthread (ShadingContext *context=NULL);
+    /// Return true if the texture handle (previously returned by
+    /// get_texture_handle()) is udim
+    virtual bool is_udim (TextureHandle *texture_handle);
 
     /// Filtered 2D texture lookup for a single point.
     ///
@@ -233,12 +217,20 @@ public:
     ///
     /// Return true if the file is found and could be opened, otherwise
     /// return false.
+    ///
+    /// If the errormessage parameter is NULL, this method is expected to
+    /// handle the errors fully, including forwarding them to the renderer
+    /// or shading system. If errormessage is non-NULL, any resulting error
+    /// messages (in case of failure, when the function returns false) will
+    /// be stored there, leaving it up to the caller/shader to handle the
+    /// error.
     virtual bool texture (ustring filename, TextureHandle *texture_handle,
                           TexturePerthread *texture_thread_info,
                           TextureOpt &options, ShaderGlobals *sg,
                           float s, float t, float dsdx, float dtdx,
                           float dsdy, float dtdy, int nchannels,
-                          float *result, float *dresultds, float *dresultdt);
+                          float *result, float *dresultds, float *dresultdt,
+                          ustring *errormessage);
 
     /// Filtered 3D texture lookup for a single point.
     ///
@@ -258,13 +250,21 @@ public:
     ///
     /// Return true if the file is found and could be opened, otherwise
     /// return false.
+    ///
+    /// If the errormessage parameter is NULL, this method is expected to
+    /// handle the errors fully, including forwarding them to the renderer
+    /// or shading system. If errormessage is non-NULL, any resulting error
+    /// messages (in case of failure, when the function returns false) will
+    /// be stored there, leaving it up to the caller/shader to handle the
+    /// error.
     virtual bool texture3d (ustring filename, TextureHandle *texture_handle,
                             TexturePerthread *texture_thread_info,
                             TextureOpt &options, ShaderGlobals *sg,
                             const Vec3 &P, const Vec3 &dPdx, const Vec3 &dPdy,
                             const Vec3 &dPdz, int nchannels,
                             float *result, float *dresultds,
-                            float *dresultdt, float *dresultdr);
+                            float *dresultdt, float *dresultdr,
+                            ustring *errormessage);
 
     /// Filtered environment lookup for a single point.
     ///
@@ -281,12 +281,20 @@ public:
     ///
     /// Return true if the file is found and could be opened, otherwise
     /// return false.
+    ///
+    /// If the errormessage parameter is NULL, this method is expected to
+    /// handle the errors fully, including forwarding them to the renderer
+    /// or shading system. If errormessage is non-NULL, any resulting error
+    /// messages (in case of failure, when the function returns false) will
+    /// be stored there, leaving it up to the caller/shader to handle the
+    /// error.
     virtual bool environment (ustring filename, TextureHandle *texture_handle,
                               TexturePerthread *texture_thread_info,
                               TextureOpt &options, ShaderGlobals *sg,
                               const Vec3 &R, const Vec3 &dRdx, const Vec3 &dRdy,
                               int nchannels, float *result,
-                              float *dresultds, float *dresultdt);
+                              float *dresultds, float *dresultdt,
+                              ustring *errormessage);
 
     /// Get information about the given texture.  Return true if found
     /// and the data has been put in *data.  Return false if the texture
@@ -301,14 +309,30 @@ public:
     /// can) use that extra information to perform a less expensive texture
     /// lookup.
     ///
-    /// Note to renderers: if sg is NULL, that means get_texture_info is
-    /// being called speculatively by the runtime optimizer, and it doesn't
-    /// know which object the shader will be run on.
-    virtual bool get_texture_info (ShaderGlobals *sg, ustring filename,
+    /// If the errormessage parameter is NULL, this method is expected to
+    /// handle the errors fully, including forwarding them to the renderer
+    /// or shading system. If errormessage is non-NULL, any resulting error
+    /// messages (in case of failure, when the function returns false) will
+    /// be stored there, leaving it up to the caller/shader to handle the
+    /// error.
+    virtual bool get_texture_info (ustring filename,
                                    TextureHandle *texture_handle,
+                                   TexturePerthread *texture_thread_info,
+                                   ShadingContext *shading_context,
                                    int subimage,
                                    ustring dataname, TypeDesc datatype,
-                                   void *data);
+                                   void *data,
+                                   ustring *errormessage);
+
+    virtual bool get_texture_info (ustring filename,
+                                   TextureHandle *texture_handle,
+                                   float s, float t,
+                                   TexturePerthread *texture_thread_info,
+                                   ShadingContext *shading_context,
+                                   int subimage,
+                                   ustring dataname, TypeDesc datatype,
+                                   void *data,
+                                   ustring *errormessage);
 
 
     /// Lookup nearest points in a point cloud. It will search for
@@ -350,6 +374,15 @@ public:
         bool shade;       ///< whether to shade what is hit
         ustring traceset; ///< named trace set
         TraceOpt () : mindist(0.0f), maxdist(1.0e30), shade(false) { }
+
+        enum class LLVMMemberIndex
+        {
+            mindist = 0,
+            maxdist,
+            shade,
+            traceset,
+            count
+        };
     };
 
     /// Immediately trace a ray from P in the direction R.  Return true
@@ -372,6 +405,29 @@ public:
     /// Return a pointer to the texture system (if available).
     virtual TextureSystem *texturesys () const;
 
+    /// Register a string with the renderer, so that the renderer can arrange
+    /// to make the string available in GPU memory as needed. Optionally specify
+    /// a variable name to associate with the string value.
+    virtual uint64_t register_string (const std::string& str,
+                                      const std::string& var_name)
+    {
+        return 0;
+    }
+
+    virtual uint64_t register_global (const std::string& var_name,
+                                      uint64_t           value)
+    {
+        return 0;
+    }
+
+    virtual bool fetch_global (const std::string& var_name,
+                               uint64_t*          value)
+    {
+        return false;
+    }
+
+
+
     /// Options we use for noise calls.
     struct NoiseOpt {
         int anisotropic;
@@ -383,7 +439,17 @@ public:
             direction(1.0f,0.0f,0.0f), bandwidth(1.0f), impulses(16.0f) { }
     };
 
+    /// A renderer may choose to support batched execution by providing pointers
+    /// to objects satisfying the BatchedRendererServices<WidthOf<#>> interface
+    /// for specific batch sizes.
+    /// Unless overridden, a nullptr is returned.
+    virtual BatchedRendererServices<16> * batched(WidthOf<16>);
+    virtual BatchedRendererServices<8> * batched(WidthOf<8>);
+
+protected:
+    TextureSystem *m_texturesys;   // A place to hold a TextureSystem
 };
 
 
+OSL_PRAGMA_WARNING_POP
 OSL_NAMESPACE_EXIT
